"""
Database interfaces.
"""
import abc
from collections import MutableSet
from collections import namedtuple


class VertexTuple(namedtuple("VertexTuple", ["label", "properties"])):
    """
    A Vertex tuple is a vertex representation which can be converted into
    a :class:`IVertex`

    :param label: Vertex label.
    :type label: :class:`str` or :obj:`None`
    :param properties: Vertex properties.
    :type properties: :class:`dict`
    """
    __slots__ = ()


# Exceptions
class EntityException(Exception):
    """
    Base container exception class.
    """


class EntitySetException(EntityException):
    """
    Base container exception class.
    """


class EntityUpdateError(EntityException):
    """
    Raised if the entity failed to update its property.
    """


class DumplicateConstraintError(EntitySetException):
    """
    Raise if you are trying to add a constraint that already exists.
    """


class DatabaseException(Exception):
    """
    Database Exception.
    """


class UnknownEntityError(DatabaseException):
    """
    Raised is the entity is unknown to the database.
    """


class ConstraintViolation(DatabaseException):
    """
    Raise if a constraint violation is found.
    """


class MultipleFoundExpectedOne(DatabaseException):
    """
    Raised when calling :meth:`.get_or_create_vertex`
    or :meth:`.get_or_create_edge` and multiple entities are found.
    """


class VertexError(DatabaseException):
    """
    Generic vertex exception.
    """


class VertexBoundByEdges(VertexError):
    """
    Raise when you are trying to remove a vertex that will is bound to
    another vertex via a edge.
    """


# Interfaces
class IGraph(object):
    """
    Interface for a property graph database.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def load(self, file_handler):
        """
        Load and import data into the database. Data should be in a JSON
        format.

        :param file_handler: A file-like object that, when read,
            produces JSON data describing a graph.
            The JSON data should be compatible with that produced by
            :meth:`~.IGraph.dump`.
        :param file_handler: :class:`file`
        """

    @abc.abstractmethod
    def dump(self, file_handler):
        """
        Export the database to a file handler.

        :param file_handler: A writable file-like object; a description of
            this graph will be written to this file encoded as JSON data
            that can be read back later with :meth:`~.IGraph.load`.
        :param file_handler: :class:`file`
        """

    @abc.abstractmethod
    def bind_to_graph(self, entity):
        """
        Bind an entity to the graph.

        :param entity: Entity that you are binding to the graph.
        :type entity: :class:`~.IEntity`
        """

    @abc.abstractmethod
    def add_vertex_constraint(self, label, key):
        """
        Add a constraint to ensure uniqueness for a particular label and
        property key.

        :param label: Vertex label which the constraint is meant for.
        :type label: :class:`str`
        :param key: Vertex property key used to ensure uniqueness.
        :type key: :class:`str`
        """

    @abc.abstractmethod
    def get_vertex_constraints(self):
        """
        Return all the known vertex constraints.

        :return: Distinct label and key pairs to
            :meth:`~.IGraph.add_vertex_constraint`.
        :rtype: Iterable of
            :class:`tuple` of label :class:`str`, key :class:`str`
        """

    @abc.abstractmethod
    def get_or_create_edge(self, head, label, tail, **kwargs):
        """
        Get or create a unique directed edge.

        .. note::

            If you wish to add in a unique undirected edge, you should add a
            directed edge in each direction.

            If ``head`` or ``tail`` is a :class:`tuple`, then
            :meth:`~.get_or_create_vertex` will always be called to create
            the vertex.


        :param head: Head vertex.
        :type head: :class:`~.IVertex` or :class:`tuple` of
            label :class:`str` and properties :class:`dict`
        :param label: Edge label.
        :type label: :class:`str`
        :param tail: Tail vertex.
        :type tail: :class:`~.IVertex` or :class:`tuple` of
            label :class:`str` and properties :class:`dict`
        :param kwargs: Property key and values to set on the new created edge.
        :type kwargs: :class:`str`, value.
        :returns: Added edge.
        :rtype: :class:`~.IEdge`
        """

    @abc.abstractmethod
    def get_or_create_vertex(self, label=None, **kwargs):
        """
        Get or create a unique vertex.

        .. note::

            Constraints will always be applied first when searching for
            vertices.

        :param label: Vertex label.
        :type label: :class:`str` or :obj:`None`
        :param kwargs: Property key and values to set on the new created
            vertex.
        :type kwargs: :class:`str`, value.
        :returns: Added vertex.
        :rtype: :class:`~.IVertex`
        """

    @abc.abstractmethod
    def add_edge(self, head, label, tail, **kwargs):
        """
        Add an directed edge to the graph.

        .. note::

            If you wish to add in a undirected edge, you should add a
            directed edge in each direction.

        :param head: Head vertex.
        :type head: :class:`~.IVertex`
        :param label: Edge label.
        :type label: :class:`str`
        :param tail: Tail vertex.
        :type tail: :class:`~.IVertex`
        :param kwargs: Property key and values to set on the new created edge.
        :type kwargs: :class:`str`, value.
        :raises ConstraintViolation: Raised if you are trying to create a
            duplicate edge between head and tail.
        :returns: Added edge.
        :rtype: :class:`~.IEdge`
        """

    @abc.abstractmethod
    def add_vertex(self, label=None, **kwargs):
        """
        Create a new vertex, add it to the graph, and return the newly
        created vertex.

        :param label: Vertex label.
        :type label: :class:`str` or :obj:`None`
        :param kwargs: Property key and values to set on the new created
            vertex.
        :type kwargs: :class:`str`, value.
        :returns: Added vertex.
        :rtype: :class:`~.IVertex`
        """

    @abc.abstractmethod
    def set_property(self, entity, **kwargs):
        """
        Set or update the entities property key and values.

        :param kwargs: Property key and values to set on the new created
            vertex.
        :type kwargs: :class:`str`, value.
        :raises ConstraintViolation: A constraint violation is raised when
            you are updating the properties of a entity and you already have
            a entity with the constrained property value.
        :raises UnknownEntityError: If you are trying to update a property
            on a :class:`~.IEntity` that is not known in the database.
        :raises TypeError: If the entity that you are trying to update is
            not supported by the database. Property updates only support
            :class:`~.Ivertex` and :class:`~.IEdge`.
        """

    @abc.abstractmethod
    def get_edge(self, id_num):
        """
        Return the edge referenced by the provided object identifier.

        :param id_num: Edge identity number.
        :type id_num: :class:`int`
        :returns: Added edge.
        :rtype: :class:`~.IEdge`
        """

    @abc.abstractmethod
    def get_edges(self, head=None, label=None, tail=None, **kwargs):
        """
        Return an iterable of all the edges in the graph that have a
        particular key/value property.

        .. note::

            See :meth:`.IEntitySet.filter` for filtering options.

        :param head: Head vertex of the edge. If :obj:`None` then
            heads will be ignored.
        :type head: :class:`~.IVertex`
        :param label: Edge label. If :obj:`None`
            then all edges will be checked for key and value.
        :type label: :class:`str` or :obj:`None`
        :param tail: Tail vertex of the edge. If :obj:`None` then
            tails will be ignored.
        :type tail: :class:`~.IVertex`
        :param kwargs: Property key and value.
        :type kwargs: :class:`str` and value.
        :returns: :class:`~.IEdge` that matched the filter criteria.
        :rtype: :class:`~.IEntitySet`
        """

    @abc.abstractmethod
    def get_vertex(self, id_num):
        """
        Return the vertex referenced by the provided object identifier.

        :param id_num: Vertex identity number.
        :type id_num: :class:`int`
        :returns: Vertex that has the identity number.
        :rtype: :class:`~.IVertex`
        """

    @abc.abstractmethod
    def get_vertices(self, label=None, **kwargs):
        """
        Return all the vertices in the graph that have a particular
        key/value property.

        .. note::

            See :meth:`.IEntitySet.filter` for filtering options.

        :param label: Vertice label. If :obj:`None`
            then all vertices will be checked for key and value.
        :param label: :class:`str` or :obj:`None`
        :param kwargs: Property key and value.
        :type kwargs: :class:`str` and value.
        :returns: :class:`~.IVertex` that matched the filter criteria.
        :rtype: :class:`~.IEntitySet`
        """

    @abc.abstractmethod
    def remove_edge(self, edge):
        """
        Remove the provided edge from the graph.

        .. note::

            Removing a edge does **not** remove the head or tail vertices,
            but only the edge between them.

        :param edge: Remove a edge/relationship.
        :type edge: :class:`~.IEdge`
        """

    @abc.abstractmethod
    def remove_vertex(self, vertex):
        """
        Remove the provided vertex from the graph.

        :param vertex: Remove a vertex/node.
        :type vertex: :class:`~.IVertex`
        :raises VertexBoundByEdges: Raised if you are trying to remove
            a vertex that is still bound or attached to another vertex
            via edge.
        """

    @abc.abstractmethod
    def close(self):
        """
        Close the instance.
        """


class IEntity(object):
    """
    Base interface for a vertex/node and edge/relationship.
    """
    __slots__ = []
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def is_bound(self):
        """
        Return True if the entity is bound to a graph.

        :returns: True is the entity is bound to a :class:`~.IGraph`
        :rtype: :class:`bool`
        """

    @abc.abstractmethod
    def remove_property(self, key):
        """
        Un-assigns a property key with its value.

        :param key: Key that you are removing.
        :type key: :class:`str`
        """

    @abc.abstractmethod
    def set_property(self, **kwargs):
        """
        Assign or update a property.

        :param kwargs: Key and value pairs.
        :type kwargs: key :class:`str` and value.
        """

    @abc.abstractmethod
    def as_dict(self):
        """
        Return the entity as a dictionary representation.

        :returns: The entity as a dictionary representation.
        :rtype: :class:`dict`
        """


class IVertex(IEntity):
    """
    Interface for a vertex/node.
    """
    __slots__ = []

    @abc.abstractmethod
    def add_in_edge(self, vertex, label=None, **kwargs):
        """
        Add and create an incoming edge between the two vertices.

        :param vertex: Edge the vertex is attached to.
        :type vertex: :class:`~.IVertex`
        :param label: Label for the edge being created.
        :type label: :class:`str`
        :param kwargs: Key and values for the edges properties.
        :type kwargs: :class:`str` and value
        """

    @abc.abstractmethod
    def add_out_edge(self, vertex, label=None, **kwargs):
        """
        Add and create an outgoing edge between the two vertices.

        :param vertex: Edge the vertex is attached to.
        :type vertex: :class:`~.IVertex`
        :param label: Label for the edge being created.
        :type label: :class:`str`
        :param kwargs: Edges property key and value pairs.
        :type kwargs: key :class:`str` and value.
        """

    @abc.abstractmethod
    def remove_edge(self, edge):
        """
        Remove a :class:`~.IEdge` from
        the vertex if it exists.

        :param edge: Edge that you are removing from the vertex.
        :type edge: :class:`~.IEdge`
        :raises KeyError: KeyError is raised if you are trying to remove
            an edge that is not found or does not exist.
        """

    @abc.abstractmethod
    def get_in_edges(self, label=None, **kwargs):
        """
        Return all the ``in`` edges to the vertex.

        :param label: Edge label.
            If :obj:`None`, all edges will be returned.
        :type label: :class:`str`
        :param kwargs: Edges property key and value pairs.
        :type kwargs: key :class:`str` and value.
        :returns: New :class:`~.IEntitySet` with filtered entities.
        :rtype: :class:`~.IEntitySet`
        """

    @abc.abstractmethod
    def get_out_edges(self, label=None, **kwargs):
        """
        Return all the ``out`` edges to the vertex.

        :param label: Edge label.
            If :obj:`None`, all edges will be returned.
        :type label: :class:`str`
        :param kwargs: Edge property key and value pairs.
        :type kwargs: key :class:`str` and value.
        :returns: New :class:`~.IEntitySet` with filtered entities.
        :rtype: :class:`~.IEntitySet`
        """

    @abc.abstractmethod
    def out_edge_count(self):
        """
        Return the total number of out edges.

        :returns: Total number of ``out`` edges.
        :rtype: :class:`int`
        """

    @abc.abstractmethod
    def in_edge_count(self):
        """
        Return the total number of in edges.

        :returns: Total number of ``in`` edges.
        :rtype: :class:`int`
        """

    @abc.abstractmethod
    def get_both_edges(self, label=None, **kwargs):
        """
        Return both ``in`` and ``out`` edges to the vertex.

        :param label: Edge label.
            If :obj:`None`, all edges will be returned.
        :type label: :class:`str`
        :param kwargs: Edge property key and value pairs.
        :type kwargs: key :class:`str` and value.
        :returns: New :class:`~.IEntitySet` with filtered entities.
        :rtype: :class:`~.IEntitySet`
        """

    @abc.abstractmethod
    def get_in_vertices(self, label=None, **kwargs):
        """
        Return the ``in`` vertices adjacent to the vertex according to the
        edge.

        :param label: Vertices label.
            If :obj:`None`, all edges will be returned.
        :type label: :class:`str`
        :param kwargs: Vertices property key and value pairs.
        :type kwargs: key :class:`str` and value.
        :returns: New :class:`~.IEntitySet` with filtered entities.
        :rtype: :class:`~.IEntitySet`
        """

    @abc.abstractmethod
    def get_out_vertices(self, label=None, **kwargs):
        """
        Return the ``out`` vertices adjacent to the vertex according to the
        edge.

        :param label: Vertices label.
            If :obj:`None`, all edges will be returned.
        :type label: :class:`str`
        :param kwargs: Vertices property key and value pairs.
        :type kwargs: key :class:`str` and value.
        :returns: New :class:`~.IEntitySet` with filtered entities.
        :rtype: :class:`~.IEntitySet`
        """

    @abc.abstractmethod
    def get_both_vertices(self, label=None, **kwargs):
        """
        Return the ``in`` and ``out`` vertices adjacent to the vertex
        according to the edges.

        :param label: Vertices label.
            If :obj:`None`, all edges will be returned.
        :type label: :class:`str`
        :param kwargs: Vertices property key and value pair.
        :type kwargs: key :class:`str` and value.
        :returns: New :class:`~.IEntitySet` with filtered entities.
        :rtype: :class:`~.IEntitySet`
        """


class IEdge(IEntity):
    """
    Interface for a edge/relationship.
    """
    __slots__ = []

    @abc.abstractmethod
    def get_in_vertex(self):
        """
        Return the ``in/head`` vertex.

        :returns: ``In`` vertex.
        :rtype: :class:`~.IVertex`
        """

    @abc.abstractmethod
    def get_out_vertex(self):
        """
        Return the ``out/tail`` vertex.

        :returns: ``Out`` vertex.
        :rtype: :class:`~.IVertex`
        """


class IEntitySet(MutableSet):
    """
    Interface for a entity containers.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.entities = set()

    def __len__(self):
        return len(self.entities)

    def __contains__(self, entity):
        return entity in self.entities

    def __iter__(self):
        return self.entities.__iter__()

    def add(self, entity):
        """
        Add a unique entity to the set.

        :param entity: Unique entity being added to the set.
        :type entity: :class:`~.IEntity`
        :raises KeyError: KeyError is raised if the entity being added
            to the set has a :attr:`~.Entity.ident` conflict with an existing
            :class:`~.IEntity`
        """
        self.entities.add(entity)

    def discard(self, entity):
        """
        Remove a entity from the current set.

        :param entity: Entity to be removed from the set.
        :type entity: :class:`~.IEntity`
        :raises KeyError: KeyError is raised if the entity being discared
            does not exists in the set.
        """
        self.entities.discard(entity)

    def remove(self, entity):
        """
        Like :meth:`.discard`, remove a entity from the current set.

        :param entity: Entity to be removed from the set.
        :type entity: :class:`~.IEntity`
        :raises KeyError: KeyError is raised if the entity being removed
            does not exists in the set.
        """
        self.discard(entity)

    @abc.abstractmethod
    def get_indexes(self):
        """
        Return all the index labels and properties.

        :returns: All the index label and property keys.
        :rtype: Iterable of :class:`tuple` of :class:`str`, :class:`str`
        """

    @abc.abstractmethod
    def get_labels(self):
        """
        Return labels known to the entity set.

        :returns: All the the labels known to the entity set.
        :rtype: Iterable of :class:`str`
        """

    @abc.abstractmethod
    def get(self, ident):
        """
        Return the :class:`.IEntity` that has the identification number
        supplied by parameter `ident`

        :param ident: Identification number.
        :type ident: :class:`int`
        :raises KeyError: Raised if there are no :class:`.IEntity` that
            has the given identification number supplied by parameter `ident`.
        :returns: The :class:`.IEntity` that has the identification number
            supplied by parameter `indent`
        :rtype: Iterable of :class:`str`
        """

    @abc.abstractmethod
    def update_index(self, entity, **kwargs):
        """
        Update the index with the new property keys.

        :param entity: Entity with a set of properties that need to be
            indexed.
        :type entity: :class:`~.IEntity`
        :param kwargs: Property key and values to set on the new created
            vertex.
        :type kwargs: :class:`str`, value.
        """

    @abc.abstractmethod
    def filter(self, label=None, **kwargs):
        """
        Filter for all entities that match the given label and properties
        returning a new
        :class:`~.IEntitySet`

        .. note::

            Keywords should be made of a property name
            (as passed to the :meth:`~.IGraph.add_vertex` or
            :meth:`~.IGraph.add_edge` methods)
            followed by one of these suffixes, to control how the given
            value is matched against the :class:`~.IEntity`'s values for
            that property.

            * __contains
            * __icontains
            * __startswith
            * __istartswith
            * __endswith
            * __iendswith
            * __le
            * __lt
            * __ge
            * __gt
            * __eq
            * __ieq
            * __ne
            * __ine

        :param label: Filter for entities that have a particular label. If
            :obj:`None`, all entities are returned.
        :type label: :class:`str`
        :param kwargs: Property key and value.
        :type kwargs: key=value
        :returns: New :class:`~.IEntitySet` with the entities that
            matched the filter criteria.
        :rtype: :class:`~.IEntitySet`
        """

    @abc.abstractmethod
    def all(self, label=None, **kwargs):
        """
        Return all the items in the container as a list.

        :param label: Filter for entities that have a particular label. If
            :obj:`None`, all entities are returned.
        :type label: :class:`str`
        :param kwargs: Property key and value.
        :type kwargs: key=value
        :returns: All the items in the container.
        :rtype: :class:`list` containing :class:`~.IEntity`
        """

    @abc.abstractmethod
    def sorted(self, key=None, reverse=False):
        """
        Sort and return all items in the container.

        :param key: Key specifies a function of one argument that is used to
            extract a comparison key from each list element. The default
            is to compare the elements directly.
        :type key: callable
        :param reverse: If set to True, then the list elements are sorted as
            if each comparison were reverted.
        :type reverse: :class:`bool`
        :returns: All the items in the container.
        :rtype: :class:`list` containing :class:`~.IEntity`
        """

    def __str__(self):
        return "<{0}> items: {1}".format(
            self.__class__.__name__, self.__len__()
        )

    def __repr__(self):  # pragma: no cover
        return self.__str__()



# pylint: disable=too-many-instance-attributes
"""
Base Test setup.
"""

import unittest2
import ruruki
from ruruki.test_utils import helpers


class TestBase(unittest2.TestCase):
    """
    Base test class.
    """
    def setUp(self):
        # First make sure that you reset the global id counters to 0
        # This is a bit of a hack
        ruruki._VID = 0  # pylint: disable=protected-access
        ruruki._EID = 0  # pylint: disable=protected-access

        self.graph = ruruki.create_graph(
            data=helpers.get_test_dump_graph_file_handler()
        )

        # See test_utils/small_people_graph.dump
        # get and set each of the vertices and edges for later use.
        self.marko = self.graph.get_vertices(name="marko").all()[0]
        self.vadas = self.graph.get_vertices(name="vadas").all()[0]
        self.lop = self.graph.get_vertices(name="lop").all()[0]
        self.josh = self.graph.get_vertices(name="josh").all()[0]
        self.ripple = self.graph.get_vertices(name="ripple").all()[0]
        self.peter = self.graph.get_vertices(name="peter").all()[0]

        self.peter_created_lop = self.graph.get_edges(
            self.peter, "created", self.lop
        ).all()[0]

        self.josh_created_lop = self.graph.get_edges(
            self.josh, "created", self.lop
        ).all()[0]

        self.marko_created_lop = self.graph.get_edges(
            self.marko, "created", tail=self.lop
        ).all()[0]

        self.marko_knows_josh = self.graph.get_edges(
            self.marko, "knows", self.josh
        ).all()[0]

        self.marko_knows_vadas = self.graph.get_edges(
            self.marko, "knows", self.vadas
        ).all()[0]

        self.josh_created_ripple = self.graph.get_edges(
            self.josh, "created", self.ripple
        ).all()[0]



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements

import json
from ruruki import create_graph
from ruruki import interfaces
from ruruki.entities import Entity, Edge, Vertex
from ruruki.test_utils import base, helpers


class TestGraph(base.TestBase):
    def test_dump(self):
        self.maxDiff = None
        tmp_file = helpers.create_tmp_file_handler()
        self.graph.dump(tmp_file)
        tmp_file.seek(0)

        # sorted the edges and vertices before comparing the two dicts
        loaded_temp = json.load(tmp_file)
        for key in loaded_temp:
            loaded_temp[key].sort()

        loaded_dump = json.load(helpers.get_test_dump_graph_file_handler())
        for key in loaded_dump:
            loaded_dump[key].sort()

        self.assertDictEqual(
            loaded_temp,
            loaded_dump,
        )

    def test_load(self):
        graph = create_graph()
        fh = helpers.get_test_dump_graph_file_handler()
        graph.load(fh)

        marko = self.graph.get_vertex(0)
        vadas = self.graph.get_vertex(1)
        lop = self.graph.get_vertex(2)
        josh = self.graph.get_vertex(3)
        ripple = self.graph.get_vertex(4)
        peter = self.graph.get_vertex(5)

        peter_created_lop = self.graph.get_edge(0)
        josh_created_lop = self.graph.get_edge(1)
        marko_created_lop = self.graph.get_edge(2)
        marko_knows_josh = self.graph.get_edge(3)
        marko_knows_vadas = self.graph.get_edge(4)
        josh_created_ripple = self.graph.get_edge(5)

        # The containers that the graph uses checks the attributes on
        # vertices and edges and not the `id` from the python built-in
        # function.
        self.assertIn(marko, self.graph)
        self.assertEqual(marko.label, "person")
        self.assertEqual(marko.ident, 0)
        self.assertDictEqual(
            marko.properties,
            {
                "age": 29,
                "name": "marko",
            },
        )

        self.assertIn(vadas, self.graph)
        self.assertEqual(vadas.label, "person")
        self.assertEqual(vadas.ident, 1)
        self.assertDictEqual(
            vadas.properties,
            {
                "age": 27,
                "name": "vadas",
            },
        )

        self.assertIn(lop, self.graph)
        self.assertEqual(lop.label, "app")
        self.assertEqual(lop.ident, 2)
        self.assertDictEqual(
            lop.properties,
            {
                "lang": "java",
                "name": "lop",
            },
        )

        self.assertIn(josh, self.graph)
        self.assertEqual(josh.label, "person")
        self.assertEqual(josh.ident, 3)
        self.assertDictEqual(
            josh.properties,
            {
                "age": 32,
                "name": "josh",
            },
        )

        self.assertIn(ripple, self.graph)
        self.assertEqual(ripple.label, "app")
        self.assertEqual(ripple.ident, 4)
        self.assertDictEqual(
            ripple.properties,
            {
                "lang": "java",
                "name": "ripple",
            },
        )

        self.assertIn(peter, self.graph)
        self.assertEqual(peter.label, "person")
        self.assertEqual(peter.ident, 5)
        self.assertDictEqual(
            peter.properties,
            {
                "age": 35,
                "name": "peter",
            },
        )

        self.assertIn(peter_created_lop, self.graph)
        self.assertEqual(
            peter_created_lop.label, "created"
        )
        self.assertEqual(
            peter_created_lop.head, peter
        )
        self.assertEqual(
            peter_created_lop.tail, lop
        )
        self.assertDictEqual(
            peter_created_lop.properties,
            {
                "weight": 0.2
            }
        )

        self.assertIn(josh_created_lop, self.graph)
        self.assertEqual(
            josh_created_lop.label, "created"
        )
        self.assertEqual(
            josh_created_lop.head, josh
        )
        self.assertEqual(
            josh_created_lop.tail, lop
        )
        self.assertDictEqual(
            josh_created_lop.properties,
            {
                "weight": 0.4
            }
        )

        self.assertIn(marko_created_lop, self.graph)
        self.assertEqual(
            marko_created_lop.label, "created"
        )
        self.assertEqual(
            marko_created_lop.head, marko
        )
        self.assertEqual(
            marko_created_lop.tail, lop
        )
        self.assertDictEqual(
            marko_created_lop.properties,
            {
                "weight": 0.4
            }
        )

        self.assertIn(marko_knows_josh, self.graph)
        self.assertEqual(
            marko_knows_josh.label, "knows"
        )
        self.assertEqual(
            marko_knows_josh.head, marko
        )
        self.assertEqual(
            marko_knows_josh.tail, josh
        )
        self.assertDictEqual(
            marko_knows_josh.properties,
            {
                "weight": 1
            }
        )

        self.assertIn(marko_knows_vadas, self.graph)
        self.assertEqual(
            marko_knows_vadas.label, "knows"
        )
        self.assertEqual(
            marko_knows_vadas.head, marko
        )
        self.assertEqual(
            marko_knows_vadas.tail, vadas
        )
        self.assertDictEqual(
            marko_knows_vadas.properties,
            {
                "weight": 0.5
            }
        )

        self.assertIn(josh_created_ripple, self.graph)
        self.assertEqual(
            josh_created_ripple.label, "created"
        )
        self.assertEqual(
            josh_created_ripple.head, josh
        )
        self.assertEqual(
            josh_created_ripple.tail, ripple
        )
        self.assertDictEqual(
            josh_created_ripple.properties,
            {
                "weight": 1
            }
        )

    def test_add_vertex(self):
        node = self.graph.add_vertex(label="NODE")
        self.assertEqual(node.label, "NODE")
        self.assertIsInstance(node, interfaces.IVertex)
        self.assertIn(node, self.graph)

    def test_add_vertex_set_id(self):
        last = max(each.ident for each in self.graph.get_vertices())
        node = self.graph.add_vertex(label="NODE")
        self.assertEqual(node.ident, last + 1)

    def test_get_vertex(self):
        self.assertEqual(self.graph.get_vertex(1), self.vadas)

    def test_get_verticies(self):
        self.assertEqual(
            self.graph.get_vertices().sorted(),
            sorted(
                [
                    self.marko, self.vadas, self.lop,
                    self.josh, self.peter, self.ripple,
                ]
            ),
        )

    def test_get_verticies_by_label(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(label="app")),
            sorted([self.lop, self.ripple]),
        )

    def test_get_verticies_by_unknown_label(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(label="UNKNOWN")),
            [],
        )

    def test_get_verticies_with_property(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(name="josh")),
            [self.josh],
        )

    def test_get_verticies_with_contains(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(name__contains="s")),
            sorted([self.vadas, self.josh]),
        )

    def test_get_verticies_where_startswith(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(name__startswith="m")),
            sorted([self.marko]),
        )

    def test_get_verticies_where_endswith(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(name__endswith="o")),
            sorted([self.marko]),
        )

    def test_remove_vertex(self):
        # First we need to remove all the edges before we can remove the
        # vertex.
        edges = self.lop.get_both_edges().all()
        for e in edges:
            self.graph.remove_edge(e)
        self.graph.remove_vertex(self.lop)
        self.assertNotIn(self.lop, self.graph)

    def test_remove_vertex_that_has_edges(self):
        self.assertRaises(
            interfaces.VertexBoundByEdges,
            self.graph.remove_vertex,
            self.lop
        )

    def test_add_edge(self):
        new_edge = self.graph.add_edge(self.marko, "FRIENDS", self.peter)
        self.assertEqual(new_edge.label, "FRIENDS")
        self.assertEqual(new_edge.head, self.marko)
        self.assertEqual(new_edge.tail, self.peter)
        self.assertIsInstance(new_edge, interfaces.IEdge)
        self.assertIn(new_edge, self.graph)

    def test_add_edge_check_id_are_set(self):
        last = max(each.ident for each in self.graph.get_edges())
        new_edge = self.graph.add_edge(self.marko, "FRIENDS", self.peter)
        self.assertEqual(new_edge.ident, last + 1)  # +1 since the last edge

    def test_get_edge(self):
        self.assertEqual(self.graph.get_edge(1), self.josh_created_lop)

    def test_get_edges(self):
        self.assertEqual(
            sorted(self.graph.get_edges()),
            sorted(
                [
                    self.peter_created_lop,
                    self.josh_created_lop,
                    self.marko_created_lop,
                    self.marko_knows_josh,
                    self.marko_knows_vadas,
                    self.josh_created_ripple,
                ]
            ),
        )

    def test_get_edges_by_with_head(self):
        self.assertEqual(
            sorted(self.graph.get_edges(head=self.marko)),
            sorted(
                [
                    self.marko_created_lop,
                    self.marko_knows_josh,
                    self.marko_knows_vadas,
                ]
            ),
        )

    def test_get_edges_by_with_tail(self):
        self.assertEqual(
            sorted(self.graph.get_edges(tail=self.lop)),
            sorted(
                [
                    self.marko_created_lop,
                    self.josh_created_lop,
                    self.peter_created_lop,
                ]
            ),
        )

    def test_get_edges_by_label(self):
        self.assertEqual(
            sorted(self.graph.get_edges(label="knows")),
            sorted([self.marko_knows_josh, self.marko_knows_vadas])
        )

    def test_get_edges_by_unknown_label(self):
        self.assertEqual(
            sorted(self.graph.get_edges(label="UNKNOWN")),
            [],
        )

    def test_get_edges_with_property(self):
        self.assertEqual(
            sorted(self.graph.get_edges(weight=0.2)),
            sorted([self.peter_created_lop]),
        )

    def test_get_edges_property_that_contains(self):
        marko_friend_josh = self.graph.get_or_create_edge(
            self.marko, "friend", self.josh,
            since="work"
        )
        self.assertEqual(
            self.graph.get_edges(since__contains="o").sorted(),
            sorted([marko_friend_josh]),
        )

    def test_get_edges_property_that_startswith(self):
        marko_friend_josh = self.graph.get_or_create_edge(
            self.marko, "friend", self.josh,
            since="work"
        )
        self.assertEqual(
            self.graph.get_edges(since__startswith="w").sorted(),
            sorted([marko_friend_josh]),
        )

    def test_get_edges_property_that_endswith(self):
        marko_friend_josh = self.graph.get_or_create_edge(
            self.marko, "friend", self.josh,
            since="work"
        )
        self.assertEqual(
            self.graph.get_edges(since__endswith="k").sorted(),
            sorted([marko_friend_josh]),
        )

    def test_remove_edge(self):
        self.graph.remove_edge(self.marko_created_lop)
        self.assertNotIn(self.marko_created_lop, self.graph)

    def test_contains_unsupported_types(self):
        unsupported = "unsupported object"
        self.assertRaises(
            TypeError,
            self.graph.__contains__,
            unsupported,
        )

    def test_contains_vertex(self):
        self.assertEqual(self.graph.__contains__(self.marko), True)

    def test_contains_vertex_not_found(self):
        v = Vertex("person", name="not_in_graph")
        self.assertEqual(self.graph.__contains__(v), False)

    def test_contains_edge(self):
        self.assertEqual(self.graph.__contains__(self.marko_knows_josh), True)

    def test_contains_edge_not_found(self):
        e = Edge(self.marko, "friend", self.josh)
        self.assertEqual(self.graph.__contains__(e), False)

    def test_set_property(self):
        self.graph.set_property(self.marko, new_prop="prop_value")

        self.assertDictEqual(
            self.marko.properties,
            {
                "name": "marko",
                "new_prop": "prop_value",
                "age": 29,
            }
        )

        self.assertEqual(
            sorted(self.graph.vertices.get_indexes()),
            sorted(
                [
                    ("app", "lang"),
                    ("app", "name"),
                    ("person", "age"),
                    ("person", "name"),
                    ("person", "new_prop"),
                ]
            ),
        )

    def test_set_property_on_an_unknown_entity(self):
        dog = Vertex("dog", name="socks")
        self.assertRaises(
            interfaces.UnknownEntityError,
            self.graph.set_property,
            dog,
            name="spot"
        )

    def test_set_property_with_empty_prop_index(self):
        dog = self.graph.get_or_create_vertex("dog", name="socks")
        self.graph.add_vertex_constraint("dog", "color")
        self.graph.set_property(dog, color="brown")
        self.assertDictEqual(
            dog.properties,
            {
                "name": "socks",
                "color": "brown",
            }
        )

    def test_set_property_indexed_entity_property(self):
        self.graph.get_or_create_vertex("dog", name="socks")
        self.graph.add_vertex_constraint("dog", "color")
        # add in another dog that would be indexed.
        self.graph.get_or_create_vertex("dog", name="patch", color="black")
        spot = self.graph.get_or_create_vertex(
            "dog", name="spot", color="white"
        )
        self.graph.set_property(spot, color="brown")
        self.assertDictEqual(
            spot.properties,
            {
                "name": "spot",
                "color": "brown",
            }
        )

    def test_entity_set_property_constraint_violation(self):
        self.graph.add_vertex_constraint("person", "surname")
        self.graph.get_or_create_vertex("person", name="John", surname="Doe")
        jane = self.graph.get_or_create_vertex("person", name="Jane")
        self.assertRaises(
            interfaces.ConstraintViolation,
            jane.set_property,
            surname="Doe",
        )

    def test_set_property_on_edge(self):
        self.graph.set_property(self.marko_knows_josh, new_prop="prop_value")

        self.assertDictEqual(
            self.marko_knows_josh.properties,
            {
                "new_prop": "prop_value",
                "weight": 1,
            }
        )

    def test_set_property_unknown_type(self):
        some_entity = Entity("SomeEntity")
        self.graph.bind_to_graph(some_entity)
        self.assertRaises(
            TypeError,
            self.graph.set_property,
            some_entity,
            name="should_not_be_set"
        )

        self.assertDictEqual(
            some_entity.properties,
            {},
        )

    def test_get_vertex_constraints(self):
        self.assertEqual(
            sorted(self.graph.get_vertex_constraints()),
            sorted(
                [
                    ("app", "name"),
                    ("person", "name"),
                ]
            ),
        )

    def test_add_vertex_constraint(self):
        current_constraints = self.graph.get_vertex_constraints()
        self.graph.add_vertex_constraint("dog", "name")
        self.assertEqual(
            sorted(self.graph.get_vertex_constraints()),
            sorted(
                [
                    ("app", "name"),
                    ("dog", "name"),
                    ("person", "name"),
                ]
            ),
        )
        self.assertNotIn(
            ("dog", "name"),
            current_constraints,
        )

    def test_bind_to_graph(self):
        sue = Vertex(100, name="Sue")
        self.graph.bind_to_graph(sue)
        self.assertIs(
            sue.graph,
            self.graph
        )


class TestGraphGetOrCreateVertices(base.TestBase):
    def test_add_new(self):
        vertices = self.graph.get_vertices().all()
        v = self.graph.get_or_create_vertex("person", name="mock_vertex")
        self.assertIn(v, self.graph)
        self.assertNotIn(v, vertices)

    def test_add_new_constraint_prop_key_not_in_kwargs(self):
        vertices = self.graph.get_vertices().all()
        self.graph.add_vertex_constraint("person", "surname")
        v = self.graph.get_or_create_vertex("person", name="mock_vertex")
        self.assertIn(v, self.graph)
        self.assertNotIn(v, vertices)

    def test_add_new_constraint_prop_key_in_kwargs(self):
        self.graph.add_vertex_constraint("person", "name")
        self.graph.add_vertex_constraint("person", "surname")
        vertices = self.graph.get_vertices().all()
        v = self.graph.get_or_create_vertex("person", surname="Marko")
        self.assertNotIn(v, vertices)

    def test_multi_add_new_no_constraints(self):
        # clear out the current constraints loaded from the dump
        self.graph._vconstraints = {}
        v = self.graph.get_or_create_vertex("person", name="mock_vertex")
        vertices = self.graph.get_vertices().all()
        dup = self.graph.get_or_create_vertex(
            "person",
            name="mock_vertex",
            age=30,
        )
        self.assertIn(dup, self.graph)
        self.assertNotIn(dup, vertices)
        self.assertNotEqual(dup, v)

    def test_multi_add_new_with_constraints(self):
        # add in a person name constraint
        self.graph.add_vertex_constraint("person", "name")
        v = self.graph.get_or_create_vertex("person", name="mock_vertex")
        vertices = self.graph.get_vertices().all()
        dup = self.graph.get_or_create_vertex(
            "person",
            name="mock_vertex",
            age=30,
        )
        self.assertIn(dup, self.graph)
        self.assertIn(dup, vertices)
        self.assertEqual(dup, v)

    def test_multi_add_new_with_constraints_unique_vertices_by_name(self):
        # add in a person name constraint
        self.graph.add_vertex_constraint("person", "name")
        v = self.graph.get_or_create_vertex(
            "person",
            name="mock_vertex",
        )

        v1 = self.graph.get_or_create_vertex(
            "person",
            name="mock_vertex1",
        )

        self.assertIn(v, self.graph)
        self.assertIn(v1, self.graph)
        self.assertNotEqual(v, v1)

    def test_more_then_one_found_with_no_constraints(self):
        self.graph.add_vertex(
            "person",
            name="mock_vertex",
            age=20
        )

        self.graph.add_vertex(
            "person",
            name="mock_vertex2",
            age=20
        )

        self.assertRaises(
            interfaces.MultipleFoundExpectedOne,
            self.graph.get_or_create_vertex,
            "person", age=20
        )

    def test_no_constraints_filter_found_one_vertex(self):
        graph = create_graph()
        graph.add_vertex("dog", name="rex")

        spot = graph.add_vertex("dog", name="spot")
        spot2 = graph.get_or_create_vertex("dog", name="spot")

        self.assertEqual(
            spot2,
            spot,
        )

    def test_no_label_no_kwargs(self):
        self.assertEqual(
            self.graph.get_or_create_vertex(),
            None
        )


class TestGraphGetOrCreateEdges(base.TestBase):
    def test_add_edge_with_head_tuple(self):
        vertices = self.graph.get_vertices("person", name="sue")
        edge = self.graph.get_or_create_edge(
            ("person", {"name": "sue"}),
            "knowns",
            self.marko,
        )

        self.assertIn(
            edge,
            self.graph
        )

        sue = self.graph.get_vertices("person", name="sue").all()[0]
        self.assertNotIn(
            sue,
            vertices,
        )

        self.assertIn(
            sue,
            self.graph,
        )

    def test_add_edge_with_tail_tuple(self):
        vertices = self.graph.get_vertices("person", name="sue")
        edge = self.graph.get_or_create_edge(
            self.marko,
            "knowns",
            ("person", {"name": "sue"}),
        )

        self.assertIn(
            edge,
            self.graph
        )

        sue = self.graph.get_vertices("person", name="sue").all()[0]
        self.assertNotIn(
            sue,
            vertices,
        )

        self.assertIn(
            sue,
            self.graph,
        )

    def test_add_edge_constraint_violation(self):
        self.assertRaises(
            interfaces.ConstraintViolation,
            self.graph.add_edge,
            self.marko,
            "knows",
            self.josh,
            place="work",
        )

    def test_edge_already_indexed(self):
        self.assertEqual(
            self.graph.get_or_create_edge(self.marko, "knows", self.josh),
            self.marko_knows_josh,
        )



"""
Some test helper functions.
"""
import tempfile
import pkg_resources


def get_test_dump_graph_file_handler():  # pylint: disable=invalid-name
    """
    Find and return the small people test graph dump file.
    """
    return open(
        pkg_resources.resource_filename(
            "ruruki", "test_utils/small_people_graph.dump"
        )
    )


def create_tmp_file_handler(content="", delete=False):
    """
    Create a temp named file.
    """
    tmp_file = tempfile.NamedTemporaryFile(delete=delete)
    tmp_file.write(content)
    tmp_file.seek(0)
    return tmp_file



"""
Graph implementations
"""
from collections import defaultdict
import json
from ruruki import interfaces
from ruruki.entities import Vertex, Edge
from ruruki.entities import EntitySet


class Graph(interfaces.IGraph):
    """
    Graph database.

    .. note::

        See :class:`~.IGraph` for doco.
    """
    def __init__(self):
        self._vconstraints = defaultdict(dict)
        self._econstraints = defaultdict()
        self.vertices = EntitySet()
        self.edges = EntitySet()

    def load(self, file_handler):
        vertex_id_mapping = {}
        data = json.load(file_handler)

        constraints = data.get("constraints", [])
        for constraint_dict in constraints:
            self.add_vertex_constraint(
                constraint_dict["label"],
                constraint_dict["key"],
            )

        vertices = sorted(data.get("vertices", []), key=lambda x: x["id"])
        for vertex_dict in vertices:
            vertex = self.get_or_create_vertex(
                vertex_dict["label"],
                **vertex_dict["properties"]
            )
            vertex_id_mapping[vertex_dict["id"]] = vertex

        edges = sorted(data.get("edges", []), key=lambda x: x["id"])
        for edge_dict in edges:
            head = vertex_id_mapping[edge_dict["head_id"]]
            tail = vertex_id_mapping[edge_dict["tail_id"]]
            self.get_or_create_edge(
                head,
                edge_dict["label"],
                tail,
                **edge_dict["properties"]
            )

    def dump(self, file_handler):
        data = {
            "vertices": [],
            "edges": [],
            "constraints": [],
        }

        for vertex in self.vertices:
            data["vertices"].append(vertex.as_dict())

        for edge in self.edges:
            data["edges"].append(edge.as_dict())

        for label, key in self.get_vertex_constraints():
            data["constraints"].append(
                {
                    "label": label,
                    "key": key
                }
            )

        json.dump(data, file_handler, indent=4, sort_keys=True)

    def add_vertex_constraint(self, label, key):
        self._vconstraints[label][key] = set()

    def get_vertex_constraints(self):
        constraints = []
        for label in self._vconstraints:
            for key in self._vconstraints[label]:
                constraints.append((label, key))
        return constraints

    def bind_to_graph(self, entity):
        entity.graph = self

    def get_or_create_vertex(self, label=None, **kwargs):
        if not label or not kwargs:
            return None

        # first check constraints.
        if label in self._vconstraints:
            for key, collection in self._vconstraints[label].items():
                if key not in kwargs:
                    continue

                for vertex in collection:
                    if vertex.properties[key] == kwargs[key]:
                        return vertex

        # no matches in constraints, so do a EntitySet filter
        vertices = self.vertices.filter(label, **kwargs).all()
        if len(vertices) > 1:
            raise interfaces.MultipleFoundExpectedOne(
                "Multiple vertices found when one expected."
            )
        elif len(vertices) == 1:
            return vertices[0]

        return self.add_vertex(label, **kwargs)

    def get_or_create_edge(self, head, label, tail, **kwargs):
        if isinstance(head, tuple):
            head = self.get_or_create_vertex(head[0], **head[1])

        if isinstance(tail, tuple):
            tail = self.get_or_create_vertex(tail[0], **tail[1])

        # There can only a single edge between head and tail with a
        # particular label. So there is not point filtering for
        # properties.
        indexed_edge = self._econstraints.get((head, label, tail))
        if indexed_edge:
            return indexed_edge
        return self.add_edge(head, label, tail, **kwargs)

    def add_edge(self, head, label, tail, **kwargs):
        if (head, label, tail) in self._econstraints:
            raise interfaces.ConstraintViolation(
                "Duplicate {0!r} edges between head {1!r} and tail {2!r} "
                "is not allowed".format(
                    label,
                    head,
                    tail
                )
            )
        edge = Edge(head, label, tail, **kwargs)
        self._econstraints[(head, label, tail)] = edge
        self.bind_to_graph(edge)
        self.edges.add(edge)
        head.out_edges.add(edge)
        tail.in_edges.add(edge)
        return edge

    def add_vertex(self, label=None, **kwargs):
        vertex = Vertex(label=label, **kwargs)

        if label in self._vconstraints:
            for key in self._vconstraints[label]:
                if key in kwargs:
                    self._vconstraints[label][key].add(vertex)

        self.bind_to_graph(vertex)
        self.vertices.add(vertex)
        return vertex

    def set_property(self, entity, **kwargs):
        if entity not in self:
            raise interfaces.UnknownEntityError(
                "Unknown entity {0!r}".format(entity)
            )

        if isinstance(entity, interfaces.IVertex):
            key_index = self._vconstraints.get(entity.label, {})
            for key, value in kwargs.items():
                if key not in key_index:
                    continue
                for indexed_entity in key_index[key]:
                    if indexed_entity != entity:
                        if indexed_entity.properties[key] == value:
                            raise interfaces.ConstraintViolation(
                                "Constraint violation with {0}".format(
                                    entity
                                )
                            )
            self.vertices.update_index(entity, **kwargs)

        if isinstance(entity, interfaces.IEdge):
            self.edges.update_index(entity, **kwargs)

        entity._update_properties(kwargs)  # pylint: disable=protected-access

    def get_edge(self, id_num):
        return self.edges.get(id_num)

    def get_vertex(self, id_num):
        return self.vertices.get(id_num)

    def get_edges(self, head=None, label=None, tail=None, **kwargs):
        if head is None and tail is None:
            return self.edges.filter(label, **kwargs)

        container = EntitySet()
        for edge in self.edges.filter(label, **kwargs):
            if head and tail is None:
                if edge.head == head:
                    container.add(edge)
            elif tail and head is None:
                if edge.tail == tail:
                    container.add(edge)
            else:
                if edge.head == head and edge.tail == tail:
                    container.add(edge)
        return container

    def get_vertices(self, label=None, **kwargs):
        return self.vertices.filter(label, **kwargs)

    def remove_edge(self, edge):
        edge.head.remove_edge(edge)
        edge.tail.remove_edge(edge)
        self.edges.remove(edge)

    def remove_vertex(self, vertex):
        if len(vertex.get_both_edges()) > 0:
            raise interfaces.VertexBoundByEdges(
                "Vertex {0!r} is still bound to another vertex "
                "by an edge. First remove all the edges on the vertex and "
                "then remove it again.".format(vertex)
            )
        self.vertices.remove(vertex)

    def close(self):  # pragma: no cover
        # Nothing to do for the close at this stage.
        return

    def __contains__(self, entity):
        is_vertex = isinstance(entity, interfaces.IVertex)
        is_edge = isinstance(entity, interfaces.IEdge)

        if not is_vertex and not is_edge:
            raise TypeError(
                "Unsupported entity type {0}".format(type(entity))
            )

        return entity in self.vertices or entity in self.edges



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

import json
import unittest2
import ruruki
from ruruki.test_utils import helpers


class TestInitHelperFunctions(unittest2.TestCase):
    def test_get_graph_plugins(self):
        self.assertEqual(
            ruruki.get_graph_plugins(),
            ["graph"]
        )

    def test_create_graph(self):
        self.assertIsInstance(
            ruruki.create_graph(),
            ruruki.interfaces.IGraph
        )

    def test_create_graph_invalid_plugin(self):
        self.assertRaises(
            ImportError,
            ruruki.create_graph,
            graph_plugin="blah"
        )

    def test_create_graph_with_data(self):
        # reset the counters for a clean run.
        ruruki._VID = 0
        ruruki._EID = 0

        data = helpers.get_test_dump_graph_file_handler()
        graph = ruruki.create_graph(data=data)

        tmp_file = helpers.create_tmp_file_handler()
        graph.dump(tmp_file)

        data.seek(0)
        tmp_file.seek(0)

        # sorted the edges and vertices before comparing the two dicts
        d1 = json.load(data)
        for key in d1:
            d1[key].sort()

        d2 = json.load(tmp_file)
        for key in d2:
            d2[key].sort()

        self.assertDictEqual(
            d1,
            d2,
        )



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods
# pylint: disable=no-member

import unittest2 as unittest
from ruruki.entities import Vertex, EntitySet, Edge
from ruruki.test_utils import base


class TestEntitySet(base.TestBase):
    def setUp(self):
        super(TestEntitySet, self).setUp()
        self.container = EntitySet([self.marko, self.josh, self.peter])

    def test_add(self):
        self.container.add(self.lop)
        self.assertEqual(
            self.container.sorted(key=lambda x: x.ident),
            sorted(
                [self.marko, self.josh, self.peter, self.lop],
                key=lambda x: x.ident,
            )
        )

    def test_add_dup_id(self):
        sue = Vertex("person", name="dup_vertex_id")
        sue.ident = 0
        self.assertRaises(
            KeyError,
            self.container.add,
            sue
        )

    def test_add_same_vertex(self):
        self.container.add(self.marko)
        self.assertEqual(
            self.container.sorted(),
            sorted([self.marko, self.josh, self.peter]),
        )

    def test_get_labels(self):
        self.container.add(self.lop)
        self.assertEqual(
            sorted(self.container.get_labels()),
            sorted(["person", "app"]),
        )

    def test_get_indexes(self):
        self.assertEqual(
            sorted(self.container.get_indexes()),
            sorted(
                [
                    ("person", "age"),
                    ("person", "name"),
                ]
            ),
        )

    def test_update_index(self):
        self.container.update_index(self.marko, surname="Foo")
        self.assertEqual(
            sorted(self.container.get_indexes()),
            sorted(
                [
                    ("person", "age"),
                    ("person", "name"),
                    ("person", "surname"),
                ]
            ),
        )

    def test_remove(self):
        self.container.remove(self.marko)
        self.assertEqual(
            self.container.sorted(),
            sorted([self.josh, self.peter])
        )

    def test_remove_unknown_entity_id(self):
        sue = Vertex(100, name="sue")
        self.assertRaises(
            KeyError,
            self.container.remove,
            sue,
        )

    def test_contains(self):
        self.assertIn(self.marko, self.container)

    def test_iter(self):
        self.assertEqual(
            self.container.sorted(),
            sorted([self.marko, self.josh, self.peter])
        )

    def test_get(self):
        self.assertEqual(self.container.get(0), self.marko)

    def test_get_no_such_entity_with_id(self):
        self.assertRaises(
            KeyError,
            self.container.get,
            100,
        )

    def test_len(self):
        self.assertEqual(3, len(self.container))

    def test_all(self):
        self.assertEqual(
            sorted(self.container.all(), key=lambda x: x.ident),
            sorted([self.marko, self.josh, self.peter], key=lambda x: x.ident)
        )

    def test_sorted(self):
        self.assertEqual(
            self.container.sorted(),
            sorted([self.marko, self.josh, self.peter])
        )

    def test_sorted_cmp_key(self):
        self.assertEqual(
            self.container.sorted(
                key=lambda x: x.properties["name"]
            ),
            sorted(
                [
                    self.marko,
                    self.josh,
                    self.peter
                ],
                key=lambda x: x.properties["name"],
            )
        )


class FilteringBase(unittest.TestCase):
    def setUp(self):
        self.marko = Vertex(
            "Father",
            name="Marko",
            surname="Jones",
            age=30,
        )
        self.john = Vertex(
            "Brother",
            name="John",
            surname="Jones",
            age=30
        )
        self.peter = Vertex(
            "Uncle",
            name="Peter",
            surname="Doe",
            age=20
        )
        self.container = EntitySet([self.marko, self.john, self.peter])


class TestFiltering(FilteringBase):
    def test_filter_for_labels_with_no_props(self):
        marko = Vertex(
            "person",
            name="Marko",
        )
        josh = Vertex(
            "person",
            name="Marko",
        )
        edge = Edge(marko, "father", josh)
        container = EntitySet()
        container.add(edge)
        self.assertEqual(
            container.filter("father").sorted(),
            sorted([edge])
        )

    def test_filter_for_vertex_with_prop_None(self):
        sue = Vertex(
            "person",
            name="sue",
            age=None,
        )
        self.container.add(sue)
        self.assertEqual(
            self.container.filter(name="Marko", age__gt=0).all(),
            [self.marko],
        )

    def test_filter_for_only_labels(self):
        self.assertEqual(
            self.container.filter("Father").sorted(),
            sorted([self.marko]),
        )

    def test_filter_with_label(self):
        self.assertEqual(
            self.container.filter("Father", name="Marko").sorted(),
            [self.marko],
        )

    def test_filter_without_label(self):
        self.assertEqual(
            self.container.filter(name="Marko").sorted(),
            [self.marko],
        )

    def test_filter_with_label_multi_props(self):
        self.assertEqual(
            self.container.filter("Father", surname="Jones", age=30).sorted(),
            [self.marko],
        )

    def test_filter_without_label_multi_props(self):
        self.assertEqual(
            self.container.filter(surname="Jones", age=30).sorted(),
            sorted([self.marko, self.john]),
        )

    def test_filter_with_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter("Father", surname="Jones", age=31).sorted(),
            [],
        )

    def test_filter_without_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(surname="Jones", age=31).sorted(),
            [],
        )

    def test_no_label_no_kwargs(self):
        self.assertEqual(
            self.container.filter(),
            self.container,
        )

    def test_unknown_prop_key_with_label(self):
        self.assertEqual(
            self.container.filter("Father", job="developer").sorted(),
            [],
        )

    def test_unknown_prop_key_without_label(self):
        self.assertEqual(
            self.container.filter(job="developer").sorted(),
            [],
        )


class TestSetOperations(FilteringBase):
    def test_union(self):
        sue = Vertex(
            "Sister",
            name="Sue"
        )
        union = self.container | EntitySet([sue, self.marko])
        self.assertEqual(
            union.sorted(),  # pylint: disable=no-member
            sorted(
                [
                    self.marko,
                    self.john,
                    self.peter,
                    sue
                ]
            )
        )

    def test_union_filter_for_labels_with_no_props(self):
        marko = Vertex(
            "person",
            name="Marko",
        )
        josh = Vertex(
            "person",
            name="Marko",
        )
        edge = Edge(marko, "father", josh)
        edge2 = Edge(josh, "son", marko)

        container1 = EntitySet()
        container2 = EntitySet()

        container1.add(edge)
        container2.add(edge2)
        container3 = container1 | container2

        self.assertEqual(
            container3.filter("father").sorted(),
            sorted([edge])
        )
        self.assertEqual(
            container3.filter("son").sorted(),
            sorted([edge2])
        )

    def test_union_with_duplicate_id(self):
        union = self.container | EntitySet([self.marko])
        self.assertEqual(
            union.sorted(),  # pylint: disable=no-member
            sorted(
                [
                    self.marko,
                    self.john,
                    self.peter,
                ]
            )
        )

    def test_intersection(self):
        sue = Vertex(
            "Sister",
            name="Sue"
        )
        intersection = self.container & EntitySet([self.marko, sue])
        self.assertEqual(
            intersection.sorted(),  # pylint: disable=no-member
            sorted(
                [
                    self.marko,
                ]
            )
        )

    def test_difference(self):
        sue = Vertex(
            "Sister",
            name="Sue"
        )
        diff = self.container - EntitySet([self.marko, sue])
        self.assertEqual(
            diff.sorted(),  # pylint: disable=no-member
            sorted(
                [
                    self.john,
                    self.peter,
                ]
            )
        )

    def test_symmetric_difference(self):
        sue = Vertex(
            "Sister",
            name="Sue"
        )
        sym_diff = self.container ^ EntitySet([self.marko, sue])
        self.assertEqual(
            sym_diff.sorted(),  # pylint: disable=no-member
            sorted(
                [
                    self.john,
                    self.peter,
                    sue,
                ]
            )
        )


class TestStartsWithFiltering(FilteringBase):
    def test_filter_with_label(self):
        self.assertEqual(
            self.container.filter("Father", name__startswith="M").sorted(),
            [self.marko],
        )

    def test_filter_without_label(self):
        self.assertEqual(
            self.container.filter(name__startswith="M").sorted(),
            [self.marko],
        )

    def test_filter_with_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                name__startswith="M",
                surname="Jones",
            ).sorted(),
            [self.marko],
        )

    def test_filter_without_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                surname__startswith="Jo",
                age=30,
            ).sorted(),
            sorted([self.marko, self.john]),
        )

    def test_filter_with_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                surname__startswith="Jo",
                age=31,
            ).sorted(),
            [],
        )

    def test_filter_without_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                surname__startswith="Jo",
                age=31,
            ).sorted(),
            [],
        )

    def test_incase_filter_with_label(self):
        self.assertEqual(
            self.container.filter("Father", name__istartswith="m").sorted(),
            [self.marko],
        )

    def test_incase_filter_without_label(self):
        self.assertEqual(
            self.container.filter(name__istartswith="m").sorted(),
            [self.marko],
        )

    def test_incase_filter_with_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                name__istartswith="m",
                surname="Jones",
            ).sorted(),
            [self.marko],
        )

    def test_incase_filter_without_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                surname__istartswith="jo",
                age=30,
            ).sorted(),
            sorted([self.marko, self.john]),
        )

    def test_incase_filter_with_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                surname__istartswith="jo",
                age=31,
            ).sorted(),
            [],
        )

    def test_incase_filter_without_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                surname__istartswith="jo",
                age=31,
            ).sorted(),
            [],
        )


class TestEndsWithFiltering(FilteringBase):
    def test_filter_with_label(self):
        self.assertEqual(
            self.container.filter("Father", name__endswith="ko").sorted(),
            [self.marko],
        )

    def test_filter_without_label(self):
        self.assertEqual(
            self.container.filter(name__endswith="ko").sorted(),
            [self.marko],
        )

    def test_filter_with_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                name__endswith="ko",
                surname="Jones",
            ).sorted(),
            [self.marko],
        )

    def test_filter_without_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                surname__endswith="s",
                age=30,
            ).sorted(),
            [self.marko, self.john],
        )

    def test_filter_with_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                surname__endswith="s",
                age=31,
            ).sorted(),
            [],
        )

    def test_filter_without_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                surname__endswith="es",
                age=31,
            ).sorted(),
            [],
        )

    def test_incase_filter_with_label(self):
        self.assertEqual(
            self.container.filter("Father", name__iendswith="ko").sorted(),
            [self.marko],
        )

    def test_incase_filter_without_label(self):
        self.assertEqual(
            self.container.filter(name__iendswith="ko").sorted(),
            [self.marko],
        )

    def test_incase_filter_with_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                name__iendswith="ko",
                surname="Jones",
            ).sorted(),
            [self.marko],
        )

    def test_incase_filter_without_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                surname__iendswith="s",
                age=30,
            ).sorted(),
            [self.marko, self.john],
        )

    def test_incase_filter_with_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                surname__iendswith="s",
                age=31,
            ).sorted(),
            [],
        )

    def test_incase_filter_without_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                surname__iendswith="es",
                age=31,
            ).sorted(),
            [],
        )


class TestContainsFiltering(FilteringBase):
    def test_filter_with_label(self):
        self.assertEqual(
            self.container.filter("Father", name__contains="k").sorted(),
            [self.marko],
        )

    def test_filter_without_label(self):
        self.assertEqual(
            self.container.filter(name__contains="k").sorted(),
            [self.marko],
        )

    def test_filter_with_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                name__contains="k",
                surname="Jones",
            ).sorted(),
            [self.marko],
        )

    def test_filter_without_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                surname__contains="o",
                age=30,
            ).sorted(),
            [self.marko, self.john],
        )

    def test_filter_with_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                surname__contains="x",
                age=31,
            ).sorted(),
            [],
        )

    def test_filter_without_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                surname__contains="x",
                age=31,
            ).sorted(),
            [],
        )

    def test_incase_filter_with_label(self):
        self.assertEqual(
            self.container.filter("Father", name__icontains="k").sorted(),
            [self.marko],
        )

    def test_incase_filter_without_label(self):
        self.assertEqual(
            self.container.filter(name__icontains="k").sorted(),
            [self.marko],
        )

    def test_incase_filter_with_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                name__icontains="k",
                surname="Jones",
            ).sorted(),
            [self.marko],
        )

    def test_incase_filter_without_label_multi_props(self):
        self.assertEqual(
            self.container.filter(
                surname__icontains="o",
                age=30,
            ).sorted(),
            [self.marko, self.john],
        )

    def test_incase_filter_with_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                "Father",
                surname__icontains="x",
                age=31,
            ).sorted(),
            [],
        )

    def test_incase_filter_without_label_not_all_props_match(self):
        self.assertEqual(
            self.container.filter(
                surname__icontains="x",
                age=31,
            ).sorted(),
            [],
        )


class TestLessEqualFiltering(FilteringBase):
    def test_filter_less_or_equal(self):
        self.assertEqual(
            self.container.filter(age__le=30).sorted(),
            sorted([self.marko, self.john, self.peter]),
        )
        self.assertEqual(
            self.container.filter(age__le=23).sorted(),
            sorted([self.peter]),
        )


class TestLessThanFiltering(FilteringBase):
    def test_filter_less_than(self):
        self.assertEqual(
            self.container.filter(age__lt=30).sorted(),
            sorted([self.peter]),
        )


class TestGreaterEqualFiltering(FilteringBase):
    def test_filter_greater_or_equal(self):
        self.assertEqual(
            self.container.filter(age__ge=30).sorted(),
            sorted([self.marko, self.john]),
        )


class TestGreaterThanFiltering(FilteringBase):
    def test_filter_greater_than(self):
        self.assertEqual(
            self.container.filter(age__gt=25).sorted(),
            sorted([self.marko, self.john]),
        )


class TestNotEqualFiltering(FilteringBase):
    def test_filter_not_equal(self):
        self.assertEqual(
            self.container.filter(name__ne="Marko").sorted(),
            sorted([self.john, self.peter]),
        )

    def test_filter_incase_not_equal(self):
        self.assertEqual(
            self.container.filter(name__ine="marko").sorted(),
            sorted([self.john, self.peter]),
        )


class TestEqualFiltering(FilteringBase):
    def test_filter_equal(self):
        self.assertEqual(
            self.container.filter(name__eq="Marko").sorted(),
            sorted([self.marko]),
        )

    def test_filter_incase_equal(self):
        self.assertEqual(
            self.container.filter(name__ieq="marko").sorted(),
            sorted([self.marko]),
        )



"""
Helpers and convenience functions for quering all the graph plugins and
for creating graph instances based on a particular graph plugin.
"""
from pkg_resources import (  # pylint: disable=no-name-in-module
    iter_entry_points,
    load_entry_point,
)

# id tracking
_EID = 0
_VID = 0


def generate_edge_id():
    """
    Generate a edge id.

    :returns: Edge id number.
    :rtype: :class:`int`
    """
    global _EID  # pylint: disable=global-statement
    ident = _EID
    _EID += 1
    return ident


def generate_vertex_id():
    """
    Generate a vertex id.

    :returns: Vertex id number.
    :rtype: :class:`int`
    """
    global _VID  # pylint: disable=global-statement
    ident = _VID
    _VID += 1
    return ident


def get_graph_plugins():
    """
    Return all the registered graph plugins available.

    :returns: All the names of the registered and available graph
        plugins.
    :rtype: Iterable of :class:`str`
    """
    return [
        each.name
        for each in iter_entry_points("ruruki.graphs")
    ]


def create_graph(graph_plugin="graph", data=None):
    """
    Create a graph based on the graph_plugin name.

    :param graph_plugin: Name of the graph plugin entry point used when
        creating the database.
    :type graph_plugin: :class:`str`
    :param data: If provided, the graph will be populated with data read
        from this file (as with graph.load()).
        Otherwise, the graph will be empty.
    :type data: :class:`file`
    :returns: A new graph instance using the implementation named in
        graph_plugin
    :rtype: :class:`~.interfaces.IGraph`
    """
    graph = load_entry_point("ruruki", "ruruki.graphs", graph_plugin)()

    if data:
        graph.load(data)

    return graph
